/*
   MRTG 2.8.12  -- Rateup
   *********************

   Rateup is a fast add-on to the great MRTG Traffic monitor.  It makes
   the database file updates much faster, and creates the graphic image
   files, ready for processing by PPMTOGIF.  It also reduces memory
   requirements by a factor of 10, and increases the speed of updates
   by a factor of at least 10.  This makes it feasible to run mrtg
   every 5 minutes.

   rateup attempts to compensate for missed updates by repeating the last
   sample, and also tries to catch bad update times.  The .log file stores
   real history every five minutes for 31 hours, then 'compresses' the 
   history into 30 minute samples for a week, then 2-hour samples for
   31 days, then daily samples for two years.  This ensures that the
   log files don't grow in size.

   The log files are a slightly different format, but convert.perl
   will fix that for you.

   Enjoy!
   Dave Rand
   dlr@bungi.com

   04/26/99 - There was some compilation bug under Watcom 10.6
   which was fixed when recompiled with VC++ 6.0
   Alexandre Steinberg
   steinberg@base.com.br
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#define GFORM_GD gdImagePng
/* BSDI 2.0.1 does not have malloc.h */
#if !defined(bsdi) && !defined(__FreeBSD__) && !defined(__OpenBSD__)
#include <malloc.h>
#endif

#include <gd.h>
#include <gdfonts.h>

#include <tagr.h>

/* Global options */

int fill_incoming_option = 1; /* Fill incoming graph */
int percent_option = 0;
int transparent_option = 1;
int zero_unknown_option = 1;

/* ... */

static char htmlf[1024];
static int histvalid;

static time_t NOW;

static char *short_si[] = {"", "k", "M", "G", "T"};
char **number_suffix = short_si;
size_t number_suffix_count = sizeof (short_si) / sizeof (short_si[0]);
static char *longup = NULL;
static char *shortup = NULL;

#define DAY_COUNT (600)		/* 400 samples is 33.33 hours */
#define DAY_SAMPLE (5*60)	/* Sample every 5 minutes */
#define WEEK_COUNT (600)	/* 400 samples is 8.33 days */
#define WEEK_SAMPLE (30*60)	/* Sample every 30 minutes */
#define MONTH_COUNT (600)	/* 400 samples is 33.33 days */
#define MONTH_SAMPLE (2*60*60)	/* Sample every 2 hours */
#define YEAR_COUNT  (2 * 366)	/* 1 sample / day, 366 days, 2 years */
#define YEAR_SAMPLE (24*60*60)	/* Sample every 24 hours */

/* One 'rounding error' per sample period, so add 4 to total and for 
   good mesure we take 10 :-) */
#define MAX_HISTORY (DAY_COUNT+WEEK_COUNT+MONTH_COUNT+YEAR_COUNT+10)

/* These are the colors used for the variouse parts of the graph */
/* the format is Red,Green,Blue */

int color_background[3]    = { 245,245,245 };
int color_light[3]    = { 194,194,194 };
int color_dark[3]     = { 100,100,100 };
int color_major[3]    = { 255,0,0 };
int color_in[3]       = { 0,235,12 };
int color_out[3]      = { 0,94,255 };
int color_grid[3]     = { 0,0,0 };
int color_in_max[3]   = { 0,166,33 };
int color_out_max[3]  = { 255,0,255 };
int color_percent[3] = { 239,159,79 };

static long kilo = (long) 1000;

#define MAXL	200		/* Maximum length of last in & out fields */


struct HISTORY
{
  time_t time;
  unsigned long in, out, percent, inmax, outmax;
}
*history;

static int Mh;

struct LAST
{
  time_t time;
  char in[MAXL], out[MAXL];
}
last;

#if !defined(__WATCOMC__) && !defined(WIN32)
# define max(a,b) ((a) > (b) ? (a) : (b))
# define min(a,b) ((a) < (b) ? (a) : (b))
#endif

/*

   notes about the NEXT macro ....

   * position n to the entry in the history array so that NOW is between
   history[n] and history[n+1]

   * calculate the interval according to steptime and position of 
   now within the history array.

   for debuging 

   fprintf (stderr,"NOW: %8lu  ST: %4lu  N: %4u HTN: %8lu HTN+1: %8lu IV: %6.0f\n", \
   now,steptime,n,history[n].time, history[n+1].time, interval);\

 */

#define NEXT(steptime)                                                        \
  inmax = outmax = avc = 0;                                                   \
  inr = outr = 0.0;                                                           \
  nextnow = now - steptime;                                                   \
  if (now == history[n].time && n < histvalid && nextnow == history[n+1].time)\
    {                                                                         \
      inr = (double) history[n].in;                                           \
      outr = (double) history[n].out;                                         \
      inmax = history[n].inmax;                                               \
      outmax = history[n].outmax;                                             \
      n++;                                                                    \
    }                                                                         \
  else if (now > history[n].time)                                             \
    logmsg(L_ERR, "trying to read ahead of the available data");              \
  else                                                                        \
    {                                                                         \
      while (now <= history[n+1].time && n < histvalid)                       \
	n++;                                                                  \
      do                                                                      \
	{                                                                     \
	  if (now >= history[n].time)                                         \
	    {                                                                 \
	      if (nextnow <= history[n+1].time)                               \
		interval = history[n].time - history[n+1].time;               \
	      else                                                            \
		interval = history[n].time - nextnow;                         \
	    }                                                                 \
	  else                                                                \
	    {                                                                 \
	      if (nextnow > history[n+1].time)                                \
		interval = steptime;                                          \
	      else                                                            \
		interval = now - history[n+1].time;                           \
	    }                                                                 \
	  inr  += (double) history[n].in * interval;                          \
	  outr += (double) history[n].out * interval;                         \
	  avc += interval;                                                    \
	  inmax =  max(inmax, history[n].inmax);                              \
	  outmax = max(outmax,history[n].outmax);                             \
	  if (nextnow <= history[n+1].time)                                   \
	    n++;                                                              \
	  else                                                                \
	    break;                                                            \
	}                                                                     \
      while (n < histvalid && nextnow < history[n].time);                     \
                                                                              \
      if (avc != steptime)                                                    \
	logmsg(L_ERR, "steptime does not match avc %8lu\n", avc);             \
                                                                              \
      inr /= avc; outr /= avc;                                                \
    }


struct image_descr
{
  unsigned long maxx;
  unsigned long maxy;
  double xscale, yscale;
  unsigned long growright, step, bits;
  int ytics;
  double yticsf;
  int peak;
  
  const char *maxin_name;
  const char *avgin_name;
  const char *curin_name;
  const char *maxout_name;
  const char *avgout_name;
  const char *curout_name;
};

int
image (const char *file, long maxvi, long maxvo,
       const struct image_descr *id,
       pp_tab_t **tab)
{
  FILE *fo;
  int i, x, y, n, type;
  long maxv;
  unsigned long origmaxvi, origmaxvo, maxs, avc, inmax, outmax;
  double inr, outr, muli, interval;
  time_t now, onow, nextnow;
  struct tm *tm;
  char *graph_label[600];
  char ylab[30];
  /* scaling helpers */
  long maxv_q;
  unsigned long valsamp, maxin, maxout, digits, maxpercent = 0;
  unsigned long sca_ten, sca_hun, nmax_q;
  double avin, avout, latestout = 0.0, latestin = 0.0, nmax, avpercent = 0.0,
    latestpercent = 0.0;
  double nex_ten, nex_hun, nex_rnd;
  double sca_max_q, dummy;
  double percent;
  char *short_si_out = "";
  struct HISTORY *lhist;

  /* ################################################# */
  /* Some general definitions for the graph generation */
    
  /* position the graph */
#define XSIZE \
    (unsigned long)((id->maxx * id->xscale) + 100 + percent_option * 30)
#define YSIZE \
    (unsigned long)((id->maxy * id->yscale) + 35)

    /* translate x/y coord into graph coord */

#define ytr(y) \
   (unsigned long)(id->maxy * id->yscale + 14 - ((y) * id->yscale))
#define xtr(x) \
   (unsigned long)(id->growright ? \
                   (id->maxx * id->xscale + 81- ((x) * id->xscale)) : \
                   (81 + ((x) * id->xscale)))
  /* ################################################# */

  /* GD LIB declarations */
  /* Declare the image */
  gdImagePtr graph, brush_out, brush_outm, brush_outp;
  /* Declare color indexes */
  int i_light, i_dark, i_background, i_major, i_in, i_out, i_grid, i_inm,
    i_outm;
  int i_outp, i_outpg;
  /* Dotted style */
  int styleDotted[3];

  /* multiplicator for bits/bytes */
  muli = 7 * id->bits + 1;
  maxvi *= muli;
  maxvo *= muli;
  origmaxvi = labs (maxvi);
  origmaxvo = labs (maxvo);
  maxv = max (maxvi, maxvo);
  if (id->step > MONTH_SAMPLE)
    {
      type = 4;
      now = (long) (NOW / YEAR_SAMPLE) * YEAR_SAMPLE;
    }
  else if (id->step > WEEK_SAMPLE)
    {
      type = 3;
      now = (long) (NOW / MONTH_SAMPLE) * MONTH_SAMPLE;
    }
  else if (id->step > DAY_SAMPLE)
    {
      type = 2;
      now = (long) (NOW / WEEK_SAMPLE) * WEEK_SAMPLE;
    }
  else
    {
      type = 1;
      now = (long) (NOW / DAY_SAMPLE) * DAY_SAMPLE;
    }
  
  if ((lhist = calloc (1, sizeof (struct HISTORY) * id->maxx)) == NULL)
    {
      logmsg(L_CRIT, "Out of memory in graph creation");
      return (1);
    }
  
  onow = now;
  avin = avout = 0.0;
  inmax = outmax = maxin = maxout = 0;
  valsamp = 0;
  for (maxs = 0, n = 0, x = 0; x < id->maxx; x++)
    {
      NEXT (id->step);

      /* scale with muli */
      inr *= muli;
      outr *= muli;
      inmax *= muli;
      outmax *= muli;

      /* ignore times when we have not sampled */
      if (inr > 0 || outr > 0)
	valsamp++;
      if (x == 0)
	{
	  latestin = inr;
	  latestout = outr;
	  if (outr)
	    latestpercent = inr * (double) 100.0 / outr;
	}
      avin += inr;
      avout += outr;
      if (id->peak)
	{
	  maxin = max (maxin, inmax);
	  maxout = max (maxout, outmax);
	  maxs = max (maxs, inmax);
	  maxs = max (maxs, outmax);
	}
      else
	{
	  maxin = max (maxin, inr);
	  maxout = max (maxout, outr);
	  maxs = max (maxs, inr);
	  maxs = max (maxs, outr);
	}
      if (percent_option && outr)
	{
	  dummy = (double) 100.0 * inr / outr;
	  maxpercent = max (dummy, maxpercent);
	}
      now -= id->step;
    }
  
  if (percent_option)
    {
      if (avout && valsamp) 
	avpercent = (double) 100. *avin / avout;
      else 
	avpercent = 0;
    }
    
  if (valsamp)
    {
      avin /= valsamp;
      avout /= valsamp;
    }
  
  add_numeric_value(tab, id->maxin_name, maxin / muli / 1024.0);
  add_numeric_value(tab, id->avgin_name, avin / muli / 1024.0);
  add_numeric_value(tab, id->curin_name, latestin / muli / 1024.0);
  add_numeric_value(tab, id->maxout_name, maxout / muli / 1024.0);
  add_numeric_value(tab, id->avgout_name, avout / muli / 1024.0);
  add_numeric_value(tab, id->curout_name, latestout / muli / 1024.0);
  
  if (maxv < 0 || maxv < maxs)
    maxv = maxs;
  now = onow;
  
  if (maxv <= 0)
    maxv = 1;
  
  /* mangle the 0.25*maxv value so, that we get a number with either */
  /* one or two digits != 0 and these digits should be at the  */
  /* start of the number ... */

  /* the ceil compensates for rounding with small numbers */
  maxv_q = ceil ((double) maxv / (double) id->ytics);	/* int */

  digits = 0;
  {
    double number = (double) maxv_q;

    while (number >= (double) kilo && digits < (number_suffix_count - 1) * 3)
      {
	number /= (double) kilo;
	digits += 3;
      }

    sca_max_q = (double) ((int) (((double) 100. * (double) number) /
				 (pow ((double) 10., (double) (int) (log10 ((double) number))))
				 + (double) 9.999) / (int) 10) / (double) 10
      * (pow ((double) 10., (double) (int) (log10 ((double) number))));
  }

  short_si_out = short_si[min (digits / 3, number_suffix_count - 1)];

  if (maxv_q)
    digits = log10 ((double) maxv_q);
  else 
    digits = 0;
    
  sca_ten = maxv_q / pow (10.0, (double) digits);
  sca_hun = maxv_q / pow (10.0, (double) digits - 1);
  nex_rnd = (sca_hun) * pow (10.0, (double) digits - 1);
  nex_ten = (sca_ten + 1) * pow (10.0, (double) digits);
  nex_hun = (sca_hun + 1) * pow (10.0, (double) digits - 1);
  
  if (nex_ten <= (1.1 * maxv_q))
    nmax_q = nex_ten;
  else if (maxv_q == nex_rnd) 
    nmax_q = nex_rnd;
  else 
    nmax_q = nex_hun;
  
  sca_max_q = nmax_q / (pow (10.0, (double) ((int) (digits / 3)) * 3));
  nmax = sca_max_q * id->ytics *
          (pow ((double) kilo, (double) ((int) (digits / 3))));

  if (nmax < 1.) 
    nmax = 1.;
    
  for (n = 0, x = 0; x < id->maxx; x++)
    {
      lhist[x].time = 0;
      graph_label[x] = NULL;
      tm = localtime (&history[n].time);
      switch (type)
	{
	default:
	case 1:
	  if (tm->tm_min < 5)
	    {
	      lhist[x].time |= 1;
	      if (tm->tm_hour == 0)
		lhist[x].time |= 2;
	    }
	  
	  if ((tm->tm_min < 5) && (tm->tm_hour % 2 == 0))
	    {
	      if ((graph_label[x] = calloc (1, sizeof (char) * 4)) == NULL)
		{
		  logmsg(L_CRIT, "Out of memory in graph labelling");
		  return (1);
		}
	      else 
		sprintf (graph_label[x], "%i", tm->tm_hour);
	    }
	  break;
	  
	case 2:
	  if (tm->tm_min < 30 && tm->tm_hour == 0)
	    {
	      lhist[x].time |= 1;
	      if (tm->tm_wday == 1)
		lhist[x].time |= 2;
	    }
	  /* fprintf(stderr,"x: %i, min:%i, hour:%i day: %i\n",
	     x,tm->tm_min,tm->tm_hour,tm->tm_wday); */
	  if ((tm->tm_min < 30) &&
	      (tm->tm_hour == 12)) {
	    if ((graph_label[x] = calloc (1, sizeof (char) * 5)) == NULL)
	      {
		logmsg(L_CRIT, "Out of memory in graph labelling");
		return (1);
	      }
	    else 
	      strftime (graph_label[x], 4, "%a", tm);
	  }
	  break;
	  
	case 3:
	  if (tm->tm_hour < 2)
	    {
	      if (tm->tm_wday == 1)
		lhist[x].time |= 1;
	      if (tm->tm_mday == 1)
		lhist[x].time |= 2;
	    }
	  /* label goes to thursday noon */
	  
	  if ((tm->tm_hour > 10) &&
	      (tm->tm_hour <= 12) &&
	      (tm->tm_wday == 4)) {
	    if ((graph_label[x] = calloc (1, sizeof (char) * 16)) == NULL)
	      {
		logmsg(L_CRIT, "Out of memory in graph labelling");
		return (1);
	      }
	    else
	      strftime (graph_label[x], 8, "Week %W", tm);
	  }
	  break;
	  
	case 4:
	  if (tm->tm_mday == 1)
	    lhist[x].time |= 1;
	  if (tm->tm_yday == 0)
	    lhist[x].time |= 2;
	  
	  if (tm->tm_mday == 15)
	    {
	      if ((graph_label[x] = calloc (1, sizeof (char) * 5)) == NULL)
		{
		  logmsg(L_CRIT, "Out of memory in graph labelling");
		  return (1);
		}
	      else 
		strftime (graph_label[x], 4, "%b", tm);
	    }
	  break;
	}
      
      NEXT (id->step);
      /*scale with muli */
      inr *= muli;
      outr *= muli;
      inmax *= muli;
      outmax *= muli;
      
      y = ((double) inr / nmax) * id->maxy;
      if (y >= id->maxy)
	y = id->maxy;
      lhist[x].in = y;
      y = ((double) outr / nmax) * id->maxy;
      if (y >= id->maxy)
	y = id->maxy;
      lhist[x].out = y;
      y = ((double) inmax / nmax) * id->maxy;
      if (y >= id->maxy)
	y = id->maxy;
      if (x && y < (lhist[x - 1].inmax))
	{
	  y += (lhist[x - 1].inmax);
	  y /= 2;
	}
      lhist[x].inmax = y;
      y = ((double) outmax / nmax) * id->maxy;
      if (y >= id->maxy)
	y = id->maxy;
      if (x && y < (lhist[x - 1].outmax))
	{
	  y += (lhist[x - 1].outmax);
	  y /= 2;
	}
      lhist[x].outmax = y;
      
      if (percent_option)
	{
	  if (outr != (unsigned long) 0)
	    percent = (double) inr / (double) outr;
	  else 
	    percent = (double) 0.;
	  if (percent > (double) 1)
	    percent = (double) 1.;
	  percent *= id->maxy;
	  lhist[x].percent = (unsigned long) percent;
	} 
      else 
	lhist[x].percent = (unsigned long) 0;
      
      now -= id->step;
    }
  
  /* the graph is made ten pixels higher to acomodate the x labels */
  graph = gdImageCreate (XSIZE, YSIZE);
  brush_out = gdImageCreate (1, 2);
  brush_outm = gdImageCreate (1, 2);
  brush_outp = gdImageCreate (1, 2);
  
  /* the first color allocated will be the background color. */
  i_background = gdImageColorAllocate (graph,
				       color_background[0],
				       color_background[1],
				       color_background[2]);
  
  i_light = gdImageColorAllocate (graph,  
				  color_light[0], color_light[1], color_light[2]);
  i_dark = gdImageColorAllocate (graph, color_dark[0], color_dark[1], color_dark[2]);
  
  if (transparent_option) 
    gdImageColorTransparent (graph, i_background);
  gdImageInterlace (graph, 1);
  
  i_major = gdImageColorAllocate (graph,
				  color_major[0],color_major[1],color_major[2]);
  i_in = gdImageColorAllocate (graph,
			       color_in[0], color_in[1], color_in[2]);
  i_out = gdImageColorAllocate (brush_out,
				color_out[0], color_out[1], color_out[2]);
  i_grid = gdImageColorAllocate (graph,
				 color_grid[0], color_grid[1], color_grid[2]);
  i_inm = gdImageColorAllocate (graph,
				color_in_max[0], color_in_max[1], color_in_max[2]);
  i_outm = gdImageColorAllocate (brush_outm,
				 color_out_max[0], color_out_max[1], color_out_max[2]);
  i_outp = gdImageColorAllocate (brush_outp,
				 color_percent[0], color_percent[1], color_percent[2]);
  i_outpg = gdImageColorAllocate (graph,
				  color_percent[0], color_percent[1], color_percent[2]);
  
  /* draw the image border */
  gdImageLine (graph, 0, 0, XSIZE - 1, 0, i_light);
  gdImageLine (graph, 1, 1, XSIZE - 2, 1, i_light);
  gdImageLine (graph, 0, 0, 0, YSIZE - 1, i_light);
  gdImageLine (graph, 1, 1, 1, YSIZE - 2, i_light);
  gdImageLine (graph, XSIZE - 1, 0, XSIZE - 1, YSIZE - 1, i_dark);
  gdImageLine (graph, 0, YSIZE - 1, XSIZE - 1, YSIZE - 1, i_dark);
  gdImageLine (graph, XSIZE - 2, 1, XSIZE - 2, YSIZE - 2, i_dark);
  gdImageLine (graph, 1, YSIZE - 2, XSIZE - 2, YSIZE - 2, i_dark);
  
  /* draw the incoming traffic */
  for (x = 0; x < id->maxx; x++)
    {
      /* peak is always above average, we therefore only draw the
	 upper part */
      if (id->peak)
	gdImageLine (graph,
		     xtr (x), ytr (lhist[x].in),
		     xtr (x), ytr (lhist[x].inmax), i_inm);
      
      if (fill_incoming_option)
	{
	  gdImageLine (graph,
		       xtr (x), ytr (0),
		       xtr (x), ytr (lhist[x].in), i_in);
	  /* draw the line a second time offset slightly. makes the graph
	   * look better if xscale is fractional */
	  gdImageLine (graph,
		       xtr (x + 0.5), ytr (0),
		       xtr (x + 0.5), ytr (lhist[x].in), i_in);
	}
      else
	gdImageLine (graph,
		     xtr (x), ytr (lhist[x].in),
		     xtr (x), ytr (lhist[x + 1].in), i_in);
    }
  
  /* draw the percentage */
  if (percent_option)
    {
      gdImageSetBrush (graph, brush_outp);
      for (x = 0; x < id->maxx - 1; x++)
	gdImageLine (graph,
		     xtr (x), ytr (lhist[x].percent),
		     xtr (x + 1), ytr (lhist[x + 1].percent),
		     gdBrushed);
    }
  
  /* draw the outgoing traffic */
  gdImageSetBrush (graph, brush_outm);
  if (id->peak)
    for (x = 0; x < id->maxx - 1; x++)
      gdImageLine (graph,
		   xtr (x), ytr (lhist[x].outmax),
		   xtr (x + 1), ytr (lhist[x + 1].outmax),
		   gdBrushed);
  
  gdImageSetBrush (graph, brush_out);
  for (x = 0; x < id->maxx - 1; x++)
    gdImageLine (graph, 
		 xtr (x), ytr (lhist[x].out),
		 xtr (x + 1), ytr (lhist[x + 1].out),
		 gdBrushed);
  
  /* draw the graph border */
  
  gdImageRectangle (graph,
		    xtr (0), ytr (0),
		    xtr (id->maxx), ytr (id->maxy), i_grid);
  
  /*create a dotted style for the grid lines */
  styleDotted[0] = i_grid;
  styleDotted[1] = gdTransparent;
  styleDotted[2] = gdTransparent;
  gdImageSetStyle (graph, styleDotted, 3);
  
  /* draw the horizontal grid */
  if ((longup == NULL) || (shortup == NULL))
    {
      if (!id->bits)
	{
	  longup = "Bytes per Second";
	  shortup = "Bytes/s";
	}
      else
	{
	  longup = "Bits per Second";
	  shortup = "Bits/s";
	}
    }
  
  if (id->maxy < gdFontSmall->w * 16)
    gdImageStringUp (graph, gdFontSmall,
		     8,
		     ytr ((id->maxy - gdFontSmall->w * strlen (shortup)) / 2),
		     (unsigned char *) shortup, i_grid);
  else 
    gdImageStringUp (graph, gdFontSmall,
		     8,
		     ytr ((id->maxy - gdFontSmall->w * strlen (longup)) / 2),
		     (unsigned char *) longup, i_grid);
  
  for (i = 0; i <= id->ytics; i++)
    {
      gdImageLine (graph,
		   xtr (-2), ytr (i * id->maxy / id->ytics),
		   xtr (1), ytr (i * id->maxy / id->ytics),
		   i_grid);
      
      gdImageLine (graph,
		   xtr (id->maxx + 2), ytr (i * id->maxy / id->ytics),
		   xtr (id->maxx - 1), ytr (i * id->maxy / id->ytics),
		   i_grid);
      
      gdImageLine (graph,
		   xtr (0), ytr (i * id->maxy / id->ytics),
		   xtr (id->maxx), ytr (i * id->maxy / id->ytics),
		   gdStyled);
      
      sprintf (ylab, "%6.1f %s", sca_max_q * i * id->yticsf, short_si_out);
      gdImageString (graph, gdFontSmall,
		     23, ytr (i * id->maxy / id->ytics + gdFontSmall->h / 2),
		     (unsigned char *) ylab, i_grid);
      
      if (percent_option)
	{
	  sprintf (ylab, "% 6.1f%%", (float) 25. * i);
	  gdImageString (graph, gdFontSmall,
			 77 + (id->maxx * id->xscale) + 1,
			 ytr (i * id->maxy / id->ytics + gdFontSmall->h / 2),
			 (unsigned char *) ylab, i_outpg);
	}
    }
  
  /* draw vertical grid and horizontal labels */
  for (x = 0; x < id->maxx; x++)
    {
      if (lhist[x].time)
	{
	  gdImageLine (graph,
		       xtr (x), ytr (-2),
		       xtr (x), ytr (1),
		       i_grid);
	  gdImageLine (graph,
		       xtr (x), ytr (0),
		       xtr (x), ytr (id->maxy),
		       gdStyled);
	}
      
      if (graph_label[x] != NULL)
	{
	  gdImageString (graph, gdFontSmall,
			 (xtr (x) - (strlen (graph_label[x]) *
				     gdFontSmall->w / 2)),
			 ytr (-4),
			 (unsigned char *) graph_label[x], i_grid);
	  free (graph_label[x]);
	}
      
      if (lhist[x].time & 2)
	gdImageLine (graph,
		     xtr (x), ytr (0),
		     xtr (x), ytr (id->maxy),
		     i_major);
    }
  
  /* draw line at peak In value in i_major color */
  /* only draw the line if it's within the graph ... */
  if (ytr (id->maxy) < ytr (((double) origmaxvi / nmax) * id->maxy))
    {
      styleDotted[0] = i_major;
      gdImageSetStyle (graph, styleDotted, 3);
      gdImageLine (graph,
		   xtr (0), ytr (((double) origmaxvi / nmax) * id->maxy),
		   xtr (id->maxx), ytr (((double) origmaxvi / nmax) * id->maxy),
		   gdStyled);
    }
  
  /* draw line at peak Out value in i_major color */
  /* only draw the line if it's within the graph ... */
  if (ytr (id->maxy) < ytr (((double) origmaxvo / nmax) * id->maxy))
    {
      styleDotted[0] = i_major;
      gdImageSetStyle (graph, styleDotted, 3);
      gdImageLine (graph,
		   xtr (0), ytr (((double) origmaxvo / nmax) * id->maxy),
		   xtr (id->maxx), ytr (((double) origmaxvo / nmax) * id->maxy),
		   gdStyled);
    }
  
  /* draw a red arrow a 0,0 */
  gdImageLine (graph, xtr (2), ytr (3), xtr (2), ytr (-3), i_major);
  gdImageLine (graph, xtr (1), ytr (3), xtr (1), ytr (-3), i_major);
  gdImageLine (graph, xtr (0), ytr (2), xtr (0), ytr (-2), i_major);
  gdImageLine (graph, xtr (-1), ytr (1), xtr (-1), ytr (-1), i_major);
  gdImageLine (graph, xtr (-2), ytr (1), xtr (-2), ytr (-1), i_major);
  gdImageLine (graph, xtr (-3), ytr (0), xtr (-3), ytr (0), i_major);
  
  if ((fo = fopen (file, "w")) == NULL)
    {
      logmsg(L_ERR, "Cannot open `%s' for writing: %s",
	     file, strerror (errno));
      exit (1);
    }
  GFORM_GD (graph, fo);
  fclose (fo);
  gdImageDestroy (graph);
  gdImageDestroy (brush_out);
  gdImageDestroy (brush_outm);
  gdImageDestroy (brush_outp);
  free (lhist);
  
  return 0;
}

double
diff (char *a, char *b)
{
  char res[MAXL], *a1, *b1, *r1;
  int c, x, m;
  
  while (!isdigit (*a))
    a++;
  while (!isdigit (*b))
    b++;
  a1 = &a[strlen (a) - 1];
  m = max (strlen (a), strlen (b));
  r1 = &res[m + 1];
  for (b1 = res; b1 <= r1; b1++)
    *b1 = ' ';
  b1 = &b[strlen (b) - 1];
  r1[1] = 0;			/* Null terminate result */
  c = 0;
  for (x = 0; x < m; x++)
    {
      if (a1 >= a && b1 >= b) 
	*r1 = ((*a1 - c) - *b1) + '0';
      else if (a1 >= a) 
	*r1 = (*a1 - c);
      else 
	*r1 = ('0' - *b1 - c) + '0';
    
      if (*r1 < '0')
	{
	  *r1 += 10;
	  c = 1;
	}
      else 
	c = 0;

      a1--;
      b1--;
      r1--;
    }
  
  if (c)
    {
      r1 = &res[m + 1];
      for (x = 0; isdigit (*r1) && x < m; x++, r1--)
	{
	  *r1 = ('9' - *r1 + c) + '0';
	  if (*r1 > '9')
	    {
	      *r1 -= 10;
	      c = 1;
	    }
	  else 
	    c = 0;
	}
      return (-atof (res));
    }
  else
    return (atof (res));
}

int
readhist (char *file)
{
  FILE *fi;
  int i, x, retcode = 0;
  char buf[256];
  struct HISTORY *hist;
  unsigned long rd[5];
  time_t cur;
  
  if ((fi = fopen (file, "r")) != NULL)
    {
      if (fscanf (fi, "%ld %s %s\n", &last.time, &last.in[0], &last.out[0]) != 3)
	{
	  logmsg (L_ERR, "%s:1: Inexpected number of fields", file);
	  retcode = 1;
	}
      cur = last.time;
      x = histvalid = 0;
      hist = history;
      while (!feof (fi))
	{
	  fgets (buf, 256, fi);
	  if (sscanf (buf, "%lu %lu %lu %lu %lu",
		      &rd[0], &rd[1], &rd[2], &rd[3], &rd[4]) < 5)
	    {
	      rd[3] = rd[1];
	      rd[4] = rd[2];
	    }
	  
	  for (i = 0; i <= 4; i++) 
	    rd[i] = rd[i] < 0 ? 0 : rd[i];

	  hist->time = rd[0];
	  hist->in = rd[1];
	  hist->out = rd[2];
	  hist->inmax = rd[3];
	  hist->outmax = rd[4];
	  if (hist->inmax < hist->in)
	    hist->inmax = hist->in;
	  if (hist->outmax < hist->out)
	    hist->outmax = hist->out;
	  if (hist->time > cur)
	    {
	      logmsg (L_WARNING, "log file %s is corrupted", file);
	      retcode = 2;
	      break;
	    }
	  
	  cur = hist->time;
	  if (++x >= Mh)
	    {
	      struct HISTORY *lh;
	      
	      Mh += MAX_HISTORY;
	      lh = realloc (history, (Mh + 1) * sizeof (struct HISTORY));
	      if (lh == NULL)
		{
		  logmsg (L_WARNING, "discarding excess data from the log file %s",
			  file);
		  break;
		}
	      hist = lh + (hist - history);
	      history = lh;
	    }
	  hist++;
	}
      histvalid = x;
      fclose (fi);
    }
  else 
    retcode = 1;

  return (retcode);
}

char * 
mkfilename (char *dir, char *name, char *suffix)
{
  int ret;
  int len = strlen (name) + strlen (suffix);
  char *buf;
  if (dir)
    len += strlen (dir) + 1;
  buf = malloc (len + 1);
  if (buf)
    {
      buf[0] = 0;
      if (dir)
	{
	  strcat (buf, dir);
	  strcat (buf, "/");
	}
      strcat (buf, name);
      strcat (buf, suffix);
    }
  return buf;
}

int
readfile (char *router)
{
  char *fname;
  int err, x;
  time_t now;
  struct HISTORY *hist;

  fname = mkfilename (NULL, router, ".log");
  err = readhist (fname);
  free (fname);
  if (err)
    {	/* Read of log file failed.  Try backup */
      fname = mkfilename (NULL, router, ".old");
      err = readhist (fname);
      free (fname);
      if (err)
	{	/* Backup failed too. New file? */
	  if (err == 2)
	    return (1);

	  /* File does not exist - it must be created */
	  now = NOW - DAY_SAMPLE;
	  hist = history;
	  histvalid = DAY_COUNT + WEEK_COUNT + MONTH_COUNT + YEAR_COUNT - 1;
	  last.time = now;
	  /* calculating a diff does not make sense */
	  last.in[0] = 'x';
	  now /= DAY_SAMPLE;
	  now *= DAY_SAMPLE;
	  for (x = 0; x < DAY_COUNT; x++, hist++)
	    {
	      hist->time = now;
	      hist->in = hist->inmax = hist->out = hist->outmax = 0;
	      now -= DAY_SAMPLE;
	    }

	  now /= WEEK_SAMPLE;
	  now *= WEEK_SAMPLE;
	  
	  for (x = 0; x < WEEK_COUNT; x++, hist++)
	    {
	      hist->time = now;
	      hist->in = hist->inmax = hist->out = hist->outmax = 0;
	      now -= WEEK_SAMPLE;
	    }
	  
	  now /= MONTH_SAMPLE;
	  now *= MONTH_SAMPLE;
	  
	  for (x = 0; x < MONTH_COUNT; x++, hist++)
	    {
	      hist->time = now;
	      hist->in = hist->inmax = hist->out = hist->outmax = 0;
	      now -= MONTH_SAMPLE;
	    }
	  
	  now /= YEAR_SAMPLE;
	  now *= YEAR_SAMPLE;
	  
	  for (x = 0; x < YEAR_COUNT; x++, hist++)
	    {
	      hist->time = now;
	      hist->in = hist->inmax = hist->out = hist->outmax = 0;
	      now -= YEAR_SAMPLE;
	    }
	}
    }
    
    return 0;
}

int
update (char *router, int in_, int out_, unsigned long abs_max, int absupdate)
{
  FILE *fp;
  char *tmpname;
  char in[32], out[32];
  time_t now, nextnow, plannow;
  long inrate, outrate;
  unsigned long avc, inmax, outmax;
  double period, interval;
  int x, n, nout;
  struct HISTORY *lhist, *hist;
  double inr, outr;
  
  sprintf (in, "%d", in_);
  sprintf (out, "%d", out_);
  
  now = NOW;
  
  if (now < last.time)
    {
      logmsg (L_WARNING, "Log file not changed");
      return 1;
    }

  tmpname = mkfilename (NULL, router, ".tmp");

  if ((lhist = calloc (1, sizeof (struct HISTORY) * (MAX_HISTORY + 1))) == NULL)
    {
      logmsg (L_CRIT, "Out of memory in update");
      return 1;
    }
  hist = lhist;
  
  period = (now - last.time);
  if (period <= 0 || period > (60 * 60) || last.in[0] == 'x')
    {	/* if last update is strange */
      if (zero_unknown_option == 1)
	{
	  inrate = 0;		/* sync unknown to zero */
	  outrate = 0;
	}
      else
	{
	  inrate = history[0].in;	/* Sync by using last value */
	  outrate = history[0].out;
	}
    }
  else
    {
      /* gauge and absolute */
      if ((absupdate != 0) && (absupdate != 3) && (absupdate != 4))
	{
	  strcpy (last.in, "0");
	  strcpy (last.out, "0");
	}
      /* gauge option */
      if (absupdate == 2)
	{
	  inrate = diff (in, last.in);
	  outrate = diff (out, last.out);
	}
      else if (absupdate == 3)
	{
	  inrate = diff (in, last.in) * (3600.0 / (period * 1.0));
	  outrate = diff (out, last.out) * (3600.0 / (period * 1.0));
	}
      else if (absupdate == 4)
	{
	  inrate = diff (in, last.in) * (60.0 / (period * 1.0));
	  outrate = diff (out, last.out) * (60.0 / (period * 1.0));
	}
      else
	{
	  inrate = diff (in, last.in) / period;
	  outrate = diff (out, last.out) / period;
	}
    }
    if (inrate < 0 || inrate > abs_max)
      {
	if (zero_unknown_option == 1) 
	  inrate = 0;		/* sync unknown to zero */
	else 
	  inrate = history[0].in;	/* Sync by using last value */
      }
    
    if (outrate < 0 || outrate > abs_max)
      {
	if (zero_unknown_option == 1) 
	    outrate = 0;	/* sync unknown to zero */
	else 
	  outrate = history[0].out;	/* Sync by using last value */
      }
    
    fp = fopen (tmpname, "w");
    if (fp)
      {
	char *oldname, *logname;
	
	fprintf (fp, "%lu %s %s\n", now, in, out);
	last.time = now;
	/* what is this good for? */
	/* gauge und absolute */
	if ((absupdate != 0) && (absupdate != 3) && (absupdate != 4))
	  {
	    strcpy (last.in, "0");
	    strcpy (last.out, "0");
	  }
	else
	  {
	    strcpy (last.in, in);
	    strcpy (last.out, out);
	  }
	
	fprintf (fp, "%lu %lu %lu %lu %lu\n",
		 now, inrate, outrate, inrate, outrate);
	
	nout = 1;
	hist->time = now;
	hist->in = inrate;
	hist->out = outrate;
	hist->inmax = inrate;
	hist->outmax = outrate;
	hist++;
	
	/* just in case we were dead for a long time, don't try to gather 
	   data from non existing log entries  */
	now = plannow = history[0].time;

	plannow /= DAY_SAMPLE;
	plannow *= DAY_SAMPLE;
	n = 0;
	
	/* gobble up every shred of data we can get ... */
	if (plannow < now)
	  {
	    NEXT ((now - plannow));
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;
	  }
	
	for (x = 1; x < DAY_COUNT; x++)
	  {
	    NEXT (DAY_SAMPLE);
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= DAY_SAMPLE;
	  }

	plannow = now;
	plannow /= WEEK_SAMPLE;
	plannow *= WEEK_SAMPLE;
	
	if (plannow < now)
	  {
	    NEXT ((now - plannow));
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;
	  }
	
	for (x = 0; x < WEEK_COUNT; x++)
	  {
	    NEXT (WEEK_SAMPLE);
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= WEEK_SAMPLE;
	  }
	
	plannow = now;
	plannow /= MONTH_SAMPLE;
	plannow *= MONTH_SAMPLE;
	
	if (plannow < now)
	  {
	    NEXT ((now - plannow));
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;
	  }
	
	for (x = 0; x < MONTH_COUNT; x++)
	  {
	    NEXT (MONTH_SAMPLE);
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= MONTH_SAMPLE;
	  }

	plannow = now;
	plannow /= YEAR_SAMPLE;
	plannow *= YEAR_SAMPLE;

	if (plannow < now)
	  {
	    NEXT ((now - plannow));
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;
	}
	
	for (x = 0; x < YEAR_COUNT; x++)
	  {
	    NEXT (YEAR_SAMPLE);
	    fprintf (fp, "%lu %lu %lu %lu %lu\n", now,
		     (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= YEAR_SAMPLE;
	  }

	if (ferror (fp))
	  {
	    logmsg (L_ERR, "Error writing new log file `%s': %s",
		    tmpname, strerror (errno));
	    free (tmpname);
	    return 1;
	  }

	fclose (fp);

	logname = mkfilename (NULL, router, ".log");
	oldname = mkfilename (NULL, router, ".old");
	unlink (oldname);
	rename (logname, oldname);
	rename (tmpname, logname);
	
	free (logname);
	free (oldname);

	for (n = 0; n < nout && n < MAX_HISTORY; n++) 
	  history[n] = lhist[n];
      }
    else
      {
	logmsg (L_ERR, "Cannot open `%s' for writing: %s",
		tmpname, strerror (errno));
	free (tmpname);
	return 1;
      }
    free (lhist);
    free (tmpname);
    
    return 0;
}

struct image_descr id_5min = {
  460, 100,    /* maxx , maxy */
  1.0, 1.0,    /* xscale , yscale */
  0, 300, 0,   /* growright, step, bits */   
  4, 1.0,      /* ytics , yticsf */
  0,           /* peak */
  
  "MAXIN", 
  "AVGIN", 
  "CURIN", 
  "MAXOUT",
  "AVGOUT",	
  "CUROUT",	
};
	
struct image_descr id_30min = {
  460, 100,    /* maxx , maxy */
  1.0, 1.0,    /* xscale , yscale */
  0, 1800, 0,  /* growright, step, bits */   
  4, 1.0,      /* ytics , yticsf */
  0,           /* peak */
  
  "HMAXIN", 
  "HAVGIN", 
  "HCURIN", 
  "HMAXOUT",
  "HAVGOUT",	
  "HCUROUT",	
};

struct image_descr id_120min = {
  460, 100,    /* maxx , maxy */
  1.0, 1.0,    /* xscale , yscale */
  0, 7200, 0,  /* growright, step, bits */   
  4, 1.0,      /* ytics , yticsf */
  0,           /* peak */
  
  "TMAXIN", 
  "TAVGIN", 
  "TCURIN", 
  "TMAXOUT",
  "TAVGOUT",	
  "TCUROUT",	
};

struct image_descr id_1440min = {
  460, 100,    /* maxx , maxy */
  1.0, 1.0,    /* xscale , yscale */
  0, 86400, 0, /* growright, step, bits */   
  4, 1.0,      /* ytics , yticsf */
  0,           /* peak */

  "DMAXIN", 
  "DAVGIN", 
  "DCURIN", 
  "DMAXOUT",
  "DAVGOUT",	
  "DCUROUT",	
};


int
frateup (int upd, char *routerpath, char *imgf, SD *d)
{
  pp_tab_t *tab = NULL;
  char *p;
  char *tabfile;
  struct tm *tm;
  int hour, minutes;
  int init;
  int host_speed_per8 = d->max / 8;
  
  if (chdir (routerpath))
    {
      logmsg (L_ERR, "chdir to `%s' failed: %s", routerpath, strerror (errno));
      return 1;
    }
	
  if ((history = calloc (1, sizeof (struct HISTORY) * (MAX_HISTORY + 1))) == NULL)
    {
      logmsg (L_CRIT, "out of memory");
      return 1;
    }

  if (readfile (d->name))
    {
      free (history);
      return 1;
    }
  
  tabfile = mkfilename (routerpath, d->name, ".tab");
  init = read_symtab (&tab, tabfile);	

  Mh = MAX_HISTORY;
  if (upd)
    {
      NOW = time (NULL);
      update (d->name, d->in / id_5min.step, d->out / id_5min.step, host_speed_per8, 2);
    }
  else
    {
      init = 1;
      NOW = last.time;
    }
  /* 5 min avg graphs */
  image (imgf, host_speed_per8, host_speed_per8, &id_5min, &tab);
  
  /* there should be 30, 120, and 24*60 min avgs */
  
  /* see if we have to update weekly, monthly or yearly graph */
  tm = localtime (&NOW);
  hour = tm->tm_hour;	
  minutes =  tm->tm_min;
  
  /* first update weekly graph ( every 30 minutes ) */
  p = mkfilename (routerpath, d->name, "-week.png");
  if (init || access (p, R_OK)
      || (minutes >= 0 && minutes < 5)
      || (minutes >= 30 && minutes < 35)) 
    image (p, host_speed_per8, host_speed_per8, &id_30min, &tab);
  free (p);
  
  /* next update monthly graph ( every 2 hours ) */
  p = mkfilename (routerpath, d->name, "-month.png");
  if (init || access (p, R_OK)
      || (minutes >= 0 && minutes < 5 && (hour % 2) == 0))
    image (p, host_speed_per8, host_speed_per8, &id_120min, &tab);
  free (p);
  
  /* and finally - yearly graph ( once per day ) */
  p = mkfilename (routerpath, d->name, "-year.png");
  if (init || access (p, R_OK)
      || (minutes >= 0 && minutes < 5 && hour == 0))
    image (p, host_speed_per8, host_speed_per8, &id_1440min, &tab);
  free (p);
  
  /* Override values of certain variables. This is based on
     the fact that add_.*() calls append the symbol to the
     beginning of the symbol table */ 
  add_string_value (&tab, "PROGRAM", PACKAGE_NAME);
  add_string_value (&tab, "VERSION", PACKAGE_VERSION);
  add_string_value (&tab, "ROUTERNAME", d->name);
  add_string_value (&tab, "ROUTERIP", d->id);
  add_string_value (&tab, "NOW", ctime (&NOW));
  
  add_numeric_value (&tab, "SPEED", d->max);
  
  /* Create html page */
  sprintf (htmlf, "%s/%s.html", routerpath, d->name);
  create_html (tab, html_template, htmlf);
  write_symtab (tab, tabfile);
  free (tabfile);
  free_tab (&tab);
  
  return 0;
}

     
int
update_graphs (SD * d)
{

  char *realdir = mkfilename (basedir, d->dir, "");
  char *pngfile = mkfilename (NULL, d->name, "-day.png");
  int rc = frateup (1, realdir, pngfile, d);
  free (realdir);
  free (pngfile);
}

int
rebuild_graphs (SD * d)
{

  char *realdir = mkfilename (basedir, d->dir, "");
  char *pngfile = mkfilename (NULL, d->name, "-day.png");
  int rc = frateup (0, realdir, pngfile, d);
  free (realdir);
  free (pngfile);
}
